@@ -24,6 +24,7 @@ limitations under the License.
 #include <X11/X.h>           // for Window, None, CopyFromParent
 #include <X11/Xatom.h>       // for XA_CARDINAL, XA_ATOM
 #include <X11/Xlib.h>        // for XEvent, XMapRaised, XSelectInput
+#include <X11/Xcms.h>        // for XcmsFailure
 #include <X11/Xutil.h>       // for XLookupString
 #include <X11/cursorfont.h>  // for XC_arrow
 #include <X11/keysym.h>      // for XK_BackSpace, XK_Tab, XK_o
@@ -45,6 +46,8 @@ limitations under the License.
 #endif
 #ifdef HAVE_XCOMPOSITE_EXT
 #include <X11/extensions/Xcomposite.h>  // for XCompositeGetOverlayWindow
+
+#include "incompatible_compositor.xbm"  // for incompatible_compositor_bits
 #endif
 #ifdef HAVE_XSCREENSAVER_EXT
 #include <X11/extensions/saver.h>      // for ScreenSaverNotify, ScreenSave...
@@ -147,6 +150,10 @@ int blank_timeout = -1;
 const char *blank_dpms_state = "off";
 //! Whether to reset the saver module when auth closes.
 int saver_reset_on_auth_close = 0;
+//! Delay we should wait before starting mapping windows to let children run.
+int saver_delay_ms = 0;
+//! Whetever stopping saver when screen is blanked.
+int saver_stop_on_blank = 0;
 
 //! The PID of a currently running notify command, or 0 if none is running.
 pid_t notify_command_pid = 0;
@@ -159,9 +166,12 @@ int blanked = 0;
 
 #ifdef HAVE_DPMS_EXT
 //! Whether DPMS needs to be disabled when unblanking. Set when blanking.
-int must_disable_dpms;
+int must_disable_dpms = 0;
 #endif
 
+//! If set by signal handler we should wake up and prompt for auth.
+static volatile sig_atomic_t signal_wakeup = 0;
+
 void ResetBlankScreenTimer(void) {
   if (blank_timeout < 0) {
     return;
@@ -191,20 +201,17 @@ void MaybeBlankScreen(Display *display) {
   }
   // Blank timer expired - blank the screen.
   blanked = 1;
-#ifdef HAVE_DPMS_EXT
-  must_disable_dpms = 0;
-#endif
   XForceScreenSaver(display, ScreenSaverActive);
   if (!strcmp(blank_dpms_state, "on")) {
     // Just X11 blanking.
-    return;
+    goto done;
   }
 #ifdef HAVE_DPMS_EXT
   // If we get here, we want to do DPMS blanking.
   int dummy;
   if (!DPMSQueryExtension(display, &dummy, &dummy)) {
     Log("DPMS is unavailable and XSECURELOCK_BLANK_DPMS_STATE not on");
-    return;
+    goto done;
   }
   CARD16 state;
   BOOL onoff;
@@ -226,12 +233,20 @@ void MaybeBlankScreen(Display *display) {
 #else
   Log("DPMS is not compiled in and XSECURELOCK_BLANK_DPMS_STATE not on");
 #endif
+done:
+  // Flush the output buffer so we turn off the display now and not a few ms
+  // later.
+  XFlush(display);
 }
 
 void ScreenNoLongerBlanked(Display *display) {
 #ifdef HAVE_DPMS_EXT
   if (must_disable_dpms) {
     DPMSDisable(display);
+    must_disable_dpms = 0;
+    // Flush the output buffer so we turn on the display now and not a
+    // few ms later. Makes our and X11's idle timer more consistent.
+    XFlush(display);
   }
 #endif
   blanked = 0;
@@ -252,6 +267,11 @@ static void HandleSIGTERM(int signo) {
   raise(signo);
 }
 
+static void HandleSIGUSR2(int unused_signo) {
+  (void)unused_signo;
+  signal_wakeup = 1;
+}
+
 enum WatchChildrenState {
   //! Request saver child.
   WATCH_CHILDREN_NORMAL,
@@ -338,7 +358,7 @@ int WakeUp(Display *dpy, Window auth_win, Window saver_win,
  *
  * This is used to prevent X11 errors from terminating XSecureLock.
  */
-int IgnoreErrorsHandler(Display *display, XErrorEvent *error) {
+int JustLogErrorsHandler(Display *display, XErrorEvent *error) {
   char buf[128];
   XGetErrorText(display, error->error_code, buf, sizeof(buf));
   buf[sizeof(buf) - 1] = 0;
@@ -346,6 +366,16 @@ int IgnoreErrorsHandler(Display *display, XErrorEvent *error) {
   return 0;
 }
 
+/*! \brief An X11 error handler that does nothing at all.
+ *
+ * This is used for calls where we expect errors to happen.
+ */
+int SilentlyIgnoreErrorsHandler(Display *display, XErrorEvent *error) {
+  (void)display;
+  (void)error;
+  return 0;
+}
+
 /*! \brief Print a version message.
  */
 void Version(void) {
@@ -412,6 +442,8 @@ void LoadDefaults() {
   blank_dpms_state = GetStringSetting("XSECURELOCK_BLANK_DPMS_STATE", "off");
   saver_reset_on_auth_close =
       GetIntSetting("XSECURELOCK_SAVER_RESET_ON_AUTH_CLOSE", 0);
+  saver_delay_ms = GetIntSetting("XSECURELOCK_SAVER_DELAY_MS", 0);
+  saver_stop_on_blank = GetIntSetting("XSECURELOCK_SAVER_STOP_ON_BLANK", 1);
 }
 
 /*! \brief Parse the command line arguments, or exit in case of failure.
@@ -425,8 +457,7 @@ void LoadDefaults() {
  * Possible errors will be printed on stderr.
  */
 void ParseArgumentsOrExit(int argc, char **argv) {
-  int i;
-  for (i = 1; i < argc; ++i) {
+  for (int i = 1; i < argc; ++i) {
     if (!strncmp(argv[i], "auth_", 5)) {
       Log("Setting auth child name from command line is DEPRECATED. Use "
           "the XSECURELOCK_AUTH environment variable instead");
@@ -802,12 +833,24 @@ int main(int argc, char **argv) {
   black.pixel = BlackPixel(display, DefaultScreen(display));
   XQueryColor(display, DefaultColormap(display, DefaultScreen(display)),
               &black);
+
+  XColor xcolor_background;
+  XColor dummy;
+  int status = XAllocNamedColor(
+      display, DefaultColormap(display, DefaultScreen(display)),
+      GetStringSetting("XSECURELOCK_BACKGROUND_COLOR", "black"),
+      &xcolor_background, &dummy);
+  unsigned long background_pixel = black.pixel;
+  if (status != XcmsFailure) {
+    background_pixel = xcolor_background.pixel;
+  }
+
   Pixmap bg = XCreateBitmapFromData(display, root_window, "\0", 1, 1);
   Cursor default_cursor = XCreateFontCursor(display, XC_arrow);
   Cursor transparent_cursor =
       XCreatePixmapCursor(display, bg, bg, &black, &black, 0, 0);
   XSetWindowAttributes coverattrs = {0};
-  coverattrs.background_pixel = black.pixel;
+  coverattrs.background_pixel = background_pixel;
   coverattrs.save_under = 1;
   coverattrs.override_redirect = 1;
   coverattrs.cursor = transparent_cursor;
@@ -855,12 +898,16 @@ int main(int argc, char **argv) {
       // off themselves in response to a full-screen window, but nevertheless
       // this is kept opt-in for now until shown reliable.
       XSetWindowAttributes obscurerattrs = coverattrs;
-      obscurerattrs.background_pixel =
-          WhitePixel(display, DefaultScreen(display));
+      obscurerattrs.background_pixmap = XCreatePixmapFromBitmapData(
+          display, root_window, incompatible_compositor_bits,
+          incompatible_compositor_width, incompatible_compositor_height,
+          BlackPixel(display, DefaultScreen(display)),
+          WhitePixel(display, DefaultScreen(display)),
+          DefaultDepth(display, DefaultScreen(display)));
       obscurer_window = XCreateWindow(
           display, root_window, 1, 1, w - 2, h - 2, 0, CopyFromParent,
           InputOutput, CopyFromParent,
-          CWBackPixel | CWSaveUnder | CWOverrideRedirect | CWCursor,
+          CWBackPixmap | CWSaveUnder | CWOverrideRedirect | CWCursor,
           &obscurerattrs);
       SetWMProperties(display, obscurer_window, "xsecurelock", "obscurer", argc,
                       argv);
@@ -963,8 +1010,8 @@ int main(int argc, char **argv) {
         XIMPreeditNone | XIMStatusNothing,     // Status might be invisible.
         XIMPreeditNone | XIMStatusNone         // Standard handling.
     };
-    size_t i;
-    for (i = 0; i < sizeof(input_styles) / sizeof(input_styles[0]); ++i) {
+    for (size_t i = 0; i < sizeof(input_styles) / sizeof(input_styles[0]);
+         ++i) {
       // Note: we draw XIM stuff in auth_window so it's above the saver/auth
       // child. However, we receive events for the grab window.
       xic = XCreateIC(xim, XNInputStyle, input_styles[i], XNClientWindow,
@@ -1000,15 +1047,29 @@ int main(int argc, char **argv) {
 
   // Acquire all grabs we need. Retry in case the window manager is still
   // holding some grabs while starting XSecureLock.
-  int retries;
   int last_normal_attempt = force_grab ? 1 : 0;
-  for (retries = 10; retries >= 0; --retries) {
+  Window previous_focused_window = None;
+  int previous_revert_focus_to = RevertToNone;
+  int retries = 10;
+  for (; retries >= 0; --retries) {
     if (AcquireGrabs(display, root_window, my_windows, n_my_windows,
                      transparent_cursor,
                      /*silent=*/retries > last_normal_attempt,
                      /*force=*/retries < last_normal_attempt)) {
       break;
     }
+    if (previous_focused_window == None) {
+      // When retrying for the first time, try to change the X11 input focus.
+      // This may close context menus and thereby allow us to grab.
+      XGetInputFocus(display, &previous_focused_window,
+                     &previous_revert_focus_to);
+      // We don't have any window mapped yet, but it should be a safe bet to set
+      // focus to the root window (most WMs allow for an easy way out from that
+      // predicament in case our restore logic fails).
+      XSetInputFocus(display, PointerRoot, RevertToPointerRoot, CurrentTime);
+      // Make this happen instantly.
+      XFlush(display);
+    }
     nanosleep(&(const struct timespec){0, 100000000L}, NULL);
   }
   if (retries < 0) {
@@ -1016,29 +1077,11 @@ int main(int argc, char **argv) {
     return EXIT_FAILURE;
   }
 
-  // Map our windows.
-  // This is done after grabbing so failure to grab does not blank the screen
-  // yet, thereby "confirming" the screen lock.
-  XMapRaised(display, background_window);
-  XClearWindow(display, background_window);  // Workaround for bad drivers.
-  XMapRaised(display, saver_window);
-  XRaiseWindow(display, auth_window);  // Don't map here.
-
-#ifdef HAVE_XCOMPOSITE_EXT
-  if (obscurer_window != None) {
-    // Map the obscurer window last so it should never become visible.
-    XMapRaised(display, obscurer_window);
-  }
-#endif
-
   if (MLOCK_PAGE(&priv, sizeof(priv)) < 0) {
     LogErrno("mlock");
     return EXIT_FAILURE;
   }
 
-  // Prevent X11 errors from killing XSecureLock. Instead, just keep going.
-  XSetErrorHandler(IgnoreErrorsHandler);
-
   struct sigaction sa;
   sigemptyset(&sa.sa_mask);
   sa.sa_flags = 0;
@@ -1046,6 +1089,10 @@ int main(int argc, char **argv) {
   if (sigaction(SIGPIPE, &sa, NULL) != 0) {
     LogErrno("sigaction(SIGPIPE)");
   }
+  sa.sa_handler = HandleSIGUSR2; // For remote wakeups by system events.
+  if (sigaction(SIGUSR2, &sa, NULL) != 0) {
+    LogErrno("sigaction(SIGUSR2)");
+  }
   sa.sa_flags = SA_RESETHAND;     // It re-raises to suicide.
   sa.sa_handler = HandleSIGTERM;  // To kill children.
   if (sigaction(SIGTERM, &sa, NULL) != 0) {
@@ -1057,7 +1104,52 @@ int main(int argc, char **argv) {
   // Need to flush the display so savers sure can access the window.
   XFlush(display);
 
+  // Figure out the initial Xss saver state. This gets updated by event.
   enum WatchChildrenState xss_requested_saver_state = WATCH_CHILDREN_NORMAL;
+#ifdef HAVE_XSCREENSAVER_EXT
+  if (scrnsaver_event_base != 0) {
+    XScreenSaverInfo *info = XScreenSaverAllocInfo();
+    XScreenSaverQueryInfo(display, root_window, info);
+    if (info->state == ScreenSaverOn && info->kind == ScreenSaverBlanked && saver_stop_on_blank) {
+      xss_requested_saver_state = WATCH_CHILDREN_SAVER_DISABLED;
+    }
+    XFree(info);
+  }
+#endif
+
+  InitBlankScreen();
+
+  XFlush(display);
+  if (WatchChildren(display, auth_window, saver_window, xss_requested_saver_state,
+                    NULL)) {
+    goto done;
+  }
+
+  // Wait for children to initialize.
+  struct timespec sleep_ts;
+  sleep_ts.tv_sec = saver_delay_ms / 1000;
+  sleep_ts.tv_nsec = (saver_delay_ms % 1000) * 1000000L;
+  nanosleep(&sleep_ts, NULL);
+
+  // Map our windows.
+  // This is done after grabbing so failure to grab does not blank the screen
+  // yet, thereby "confirming" the screen lock.
+  XMapRaised(display, saver_window);
+  XMapRaised(display, background_window);
+  XClearWindow(display, background_window);  // Workaround for bad drivers.
+  XRaiseWindow(display, auth_window);  // Don't map here.
+
+#ifdef HAVE_XCOMPOSITE_EXT
+  if (obscurer_window != None) {
+    // Map the obscurer window last so it should never become visible.
+    XMapRaised(display, obscurer_window);
+  }
+#endif
+  XFlush(display);
+
+  // Prevent X11 errors from killing XSecureLock. Instead, just keep going.
+  XSetErrorHandler(JustLogErrorsHandler);
+
   int x11_fd = ConnectionNumber(display);
 
   if (x11_fd == xss_sleep_lock_fd && xss_sleep_lock_fd != -1) {
@@ -1066,8 +1158,6 @@ int main(int argc, char **argv) {
     xss_sleep_lock_fd = -1;
   }
 
-  InitBlankScreen();
-
   int background_window_mapped = 0, background_window_visible = 0,
       auth_window_mapped = 0, saver_window_mapped = 0,
       need_to_reinstate_grabs = 0, xss_lock_notified = 0;
@@ -1084,7 +1174,7 @@ int main(int argc, char **argv) {
 
     // Make sure to shut down the saver when blanked. Saves power.
     enum WatchChildrenState requested_saver_state =
-        blanked ? WATCH_CHILDREN_SAVER_DISABLED : xss_requested_saver_state;
+      (saver_stop_on_blank && blanked) ? WATCH_CHILDREN_SAVER_DISABLED : xss_requested_saver_state;
 
     // Now check status of our children.
     if (WatchChildren(display, auth_window, saver_window, requested_saver_state,
@@ -1129,6 +1219,22 @@ int main(int argc, char **argv) {
       // Otherwise, we're still alive. Re-check next time.
     }
 
+    if (signal_wakeup) {
+      // A signal was received to request a wakeup. Clear the flag
+      // and proceed to auth. Technically this check involves a race
+      // condition between check and set since we don't block signals,
+      // but this should not make a difference since screen wakeup is
+      // idempotent.
+      signal_wakeup = 0;
+#ifdef DEBUG_EVENTS
+      Log("WakeUp on signal");
+#endif
+      UnblankScreen(display);
+      if (WakeUp(display, auth_window, saver_window, NULL)) {
+        goto done;
+      }
+    }
+
     // Handle all events.
     while (XPending(display) && (XNextEvent(display, &priv.ev), 1)) {
       if (XFilterEvent(&priv.ev, None)) {
@@ -1272,7 +1378,8 @@ int main(int argc, char **argv) {
             // Map Ctrl-Backspace to Ctrl-U (clear entry line).
             priv.buf[0] = '\025';
             priv.buf[1] = 0;
-          } else if (have_switch_user_command && priv.keysym == XK_o &&
+          } else if (have_switch_user_command &&
+                     (priv.keysym == XK_o || priv.keysym == XK_0) &&
                      (((priv.ev.xkey.state & ControlMask) &&
                        (priv.ev.xkey.state & Mod1Mask)) ||
                       (priv.ev.xkey.state & Mod4Mask))) {
@@ -1469,14 +1576,17 @@ done:
   // Make sure no DPMS changes persist.
   UnblankScreen(display);
 
+  if (previous_focused_window != None) {
+    XSetErrorHandler(SilentlyIgnoreErrorsHandler);
+    XSetInputFocus(display, previous_focused_window, previous_revert_focus_to,
+                   CurrentTime);
+    XSetErrorHandler(JustLogErrorsHandler);
+  }
+
   // Wipe the password.
   explicit_bzero(&priv, sizeof(priv));
 
   // Free our resources, and exit.
-  XDestroyWindow(display, auth_window);
-  XDestroyWindow(display, saver_window);
-  XDestroyWindow(display, background_window);
-
 #ifdef HAVE_XCOMPOSITE_EXT
   if (obscurer_window != None) {
     // Destroy the obscurer window first so it should never become visible.
@@ -1486,6 +1596,9 @@ done:
     XCompositeReleaseOverlayWindow(display, composite_window);
   }
 #endif
+  XDestroyWindow(display, auth_window);
+  XDestroyWindow(display, saver_window);
+  XDestroyWindow(display, background_window);
 
   XFreeCursor(display, transparent_cursor);
   XFreeCursor(display, default_cursor);
